"""
Test cases for ldaptor.protocols.ldap.ldapfilter module.
"""

from twisted.trial import unittest
from ldaptor.protocols import pureldap
from ldaptor import ldapfilter


class RFC2254Examples(unittest.TestCase):
    def test_cn(self):
        text = '(cn=Babs Jensen)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='Babs Jensen'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_not_cn(self):
        text = '(!(cn=Tim Howes))'
        filt = pureldap.LDAPFilter_not(
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='Tim Howes')))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_and_or(self):
        text = '(&(objectClass=Person)(|(sn=Jensen)(cn=Babs J*)))'
        filt = pureldap.LDAPFilter_and(
            [ pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='objectClass'),
            assertionValue=pureldap.LDAPAssertionValue(value='Person')),
              pureldap.LDAPFilter_or([ pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='sn'),
            assertionValue=pureldap.LDAPAssertionValue(value='Jensen')),
                                       pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[ pureldap.LDAPFilter_substrings_initial(value='Babs J')
                         ])
                                       ]),
              ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_substrings(self):
        text = '(o=univ*of*mich*)'
        filt = pureldap.LDAPFilter_substrings(
            type='o',
            substrings=[ pureldap.LDAPFilter_substrings_initial(value='univ'),
                         pureldap.LDAPFilter_substrings_any(value='of'),
                         pureldap.LDAPFilter_substrings_any(value='mich'),
                         ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)


    def test_extensible_1(self):
        text = '(cn:1.2.3.4.5:=Fred Flintstone)'
        filt = pureldap.LDAPFilter_extensibleMatch(
            type='cn',
            dnAttributes=False,
            matchingRule='1.2.3.4.5',
            matchValue='Fred Flintstone',
            )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_2(self):
        text = '(sn:dn:2.4.6.8.10:=Barney Rubble)'
        filt = pureldap.LDAPFilter_extensibleMatch(
            type='sn',
            dnAttributes=True,
            matchingRule='2.4.6.8.10',
            matchValue='Barney Rubble',
            )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_3(self):
        text = '(o:dn:=Ace Industry)'
        filt = pureldap.LDAPFilter_extensibleMatch(
            type='o',
            dnAttributes=True,
            matchingRule=None,
            matchValue='Ace Industry',
            )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_4(self):
        text = '(:dn:2.4.6.8.10:=Dino)'
        filt = pureldap.LDAPFilter_extensibleMatch(
            type=None,
            dnAttributes=True,
            matchingRule='2.4.6.8.10',
            matchValue='Dino',
            )
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_extensible_5(self):
        text = '(cn:1.2.3.4.5:=Fred Flintstone)'
        filt = pureldap.LDAPFilter_extensibleMatch(
            type='cn',
            dnAttributes=None,
            matchingRule='1.2.3.4.5',
            matchValue='Fred Flintstone',
            )
        self.assertEqual(filt.asText(), text)

    def test_escape_parens(self):
        text = r'(o=Parens R Us \28for all your parenthetical needs\29)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='o'),
            assertionValue=pureldap.LDAPAssertionValue(value='Parens R Us (for all your parenthetical needs)'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_escape_asterisk(self):
        text = r'(cn=*\2A*)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[ pureldap.LDAPFilter_substrings_any(value='*'),
                         ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text.lower())

    def test_escape_backslash(self):
        text = r'(filename=C:\5cMyFile)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='filename'),
            assertionValue=pureldap.LDAPAssertionValue(value=r'C:\MyFile'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_escape_binary(self):
        text = r'(bin=\00\00\00\04)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='bin'),
            assertionValue=pureldap.LDAPAssertionValue(value='\00\00\00\04'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)

    def test_escape_utf8(self):
        text = r'(sn=Lu\c4\8di\c4\87)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='sn'),
            assertionValue=pureldap.LDAPAssertionValue(value='Lu\xc4\x8di\xc4\x87'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        #self.assertEqual(filt.asText(), text)

class TestValid(unittest.TestCase):
    def test_item_present(self):
        text = r'(cn=*)'
        filt = pureldap.LDAPFilter_present(value='cn')
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_simple(self):
        text = r'(cn=foo)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_init(self):
        text = r'(cn=foo*)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_final(self):
        text = r'(cn=*foo)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_final('foo'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_any(self):
        text = r'(cn=*foo*)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_aa(self):
        text = r'(cn=*foo*bar*)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_ia(self):
        text = r'(cn=foo*bar*)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_iaa(self):
        text = r'(cn=foo*bar*baz*)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_any('baz'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_if(self):
        text = r'(cn=foo*bar)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_final('bar'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_iaf(self):
        text = r'(cn=foo*bar*baz)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_final('baz'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_iaaf(self):
        text = r'(cn=foo*bar*baz*quux)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_any('baz'),
                        pureldap.LDAPFilter_substrings_final('quux'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_af(self):
        text = r'(cn=*foo*bar)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        pureldap.LDAPFilter_substrings_final('bar'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_item_substring_aaf(self):
        text = r'(cn=*foo*bar*baz)'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_final('baz'),
                        ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_not_item(self):
        text = r'(!(cn=foo))'
        filt = pureldap.LDAPFilter_not(
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo')))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_or_item(self):
        text = r'(|(cn=foo)(cn=bar))'
        filt = pureldap.LDAPFilter_or([
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo')),
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='bar')),
            ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_and_item(self):
        text = r'(&(cn=foo)(cn=bar))'
        filt = pureldap.LDAPFilter_and([
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo')),
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='bar')),
            ])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_andornot(self):
        text = r'(&(!(|(cn=foo)(cn=bar)))(sn=a*b*c*d))'
        filt = pureldap.LDAPFilter_and([
            pureldap.LDAPFilter_not(
            pureldap.LDAPFilter_or([
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo')),
            pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='bar')),
            ])),
            pureldap.LDAPFilter_substrings(
            type='sn',
            substrings=[pureldap.LDAPFilter_substrings_initial('a'),
                        pureldap.LDAPFilter_substrings_any('b'),
                        pureldap.LDAPFilter_substrings_any('c'),
                        pureldap.LDAPFilter_substrings_final('d'),
                        ])])
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

    def test_whitespace_beforeCloseParen(self):
        text = r'(cn=foo )'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo '))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)


    def test_whitespace_afterEq(self):
        text = r'(cn= foo)'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value=' foo'))
        self.assertEqual(ldapfilter.parseFilter(text), filt)
        self.assertEqual(filt.asText(), text)

class TestInvalid(unittest.TestCase):
    def test_closeParen_1(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(&(|(mail=)@*)(uid=)))(mail=*))')

    def test_closeParen_2(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(|(mail=)@*)(uid=)))')

    def test_closeParen_3(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(mail=)@*)')

    def test_closeParen_4(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(uid=))')


    def test_openParen_1(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(&(|(mail=(@*)(uid=())(mail=*))')

    def test_openParen_2(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(|(mail=(@*)(uid=())')

    def test_openParen_3(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(mail=(@*)')

    def test_openParen_4(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          '(uid=()')

    def test_whitespace_leading(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          r' (cn=foo)')

    def test_whitespace_trailing(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          r'(cn=foo) ')

    def test_whitespace_afterOpenParen(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          r'( cn=foo)')

    def test_whitespace_beforeEq(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          r'(cn =foo)')

class TestMaybeSubstring(unittest.TestCase):
    def test_item_present(self):
        text = r'*'
        filt = pureldap.LDAPFilter_present(value='cn')
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_simple(self):
        text = r'foo'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='foo'))
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_init(self):
        text = r'foo*'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_final(self):
        text = r'*foo'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_final('foo'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_any(self):
        text = r'*foo*'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_aa(self):
        text = r'*foo*bar*'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_ia(self):
        text = r'foo*bar*'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_iaa(self):
        text = r'foo*bar*baz*'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_any('baz'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_if(self):
        text = r'foo*bar'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_final('bar'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_iaf(self):
        text = r'foo*bar*baz'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_final('baz'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_iaaf(self):
        text = r'foo*bar*baz*quux'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_initial('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_any('baz'),
                        pureldap.LDAPFilter_substrings_final('quux'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_af(self):
        text = r'*foo*bar'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        pureldap.LDAPFilter_substrings_final('bar'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_item_substring_aaf(self):
        text = r'*foo*bar*baz'
        filt = pureldap.LDAPFilter_substrings(
            type='cn',
            substrings=[pureldap.LDAPFilter_substrings_any('foo'),
                        pureldap.LDAPFilter_substrings_any('bar'),
                        pureldap.LDAPFilter_substrings_final('baz'),
                        ])
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

    def test_escape_simple(self):
        text = r'f\2aoo(bar'
        filt = pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value='cn'),
            assertionValue=pureldap.LDAPAssertionValue(value='f*oo(bar'))
        self.assertEqual(ldapfilter.parseMaybeSubstring('cn', text), filt)

class TestWhitespace(unittest.TestCase):
    def test_escape(self):
        self.assertRaises(ldapfilter.InvalidLDAPFilter,
                          ldapfilter.parseFilter,
                          r'(cn=\ 61)')



"""
Test cases for the ldaptor.samba.smbpassword module.
"""

from twisted.trial import unittest
from ldaptor.samba import smbpassword
from ldaptor import config

class TestNTHash(unittest.TestCase):
    """
    Unit tests for smbpassword.nthash.
    """
    knownValues=[ # password, expected_result
        (b'', b'31D6CFE0D16AE931B73C59D7E0C089C0'),
        (b'foo', b'AC8E657F83DF82BEEA5D43BDAF7800CC'),
        (127 * b'x', b'25900FAB94A048BCF438615217776562'),
        (128 * b'x', b'65681023D0CB5E7E96FF662150EF060D'),
        (129 * b'x', b'65681023D0CB5E7E96FF662150EF060D'),
        (1000 * b'x', b'65681023D0CB5E7E96FF662150EF060D'),
    ]

    def testKnownValues(self):
        """nthash(...) gives known results"""
        self.assertEqual(
            self.knownValues,
            [(p, smbpassword.nthash(p)) for p, _ in self.knownValues]
        )


class TestLMHash(unittest.TestCase):
    """
    Unit tests for smbpassword.lmhash.
    """
    knownValues=[ # password, expected_result
        (b'', b'AAD3B435B51404EEAAD3B435B51404EE'),
        (b'foo', b'5BFAFBEBFB6A0942AAD3B435B51404EE'),
        (13 * b'x', b'3AA62DBBEFDB676366B4159AF5A7C45C'),
        (14 * b'x', b'3AA62DBBEFDB67633AA62DBBEFDB6763'),
        (15 * b'x', b'3AA62DBBEFDB67633AA62DBBEFDB6763'),
        (100 * b'x', b'3AA62DBBEFDB67633AA62DBBEFDB6763'),
        (b'1234567abcdefg', b'0182BD0BD4444BF8E0C510199CC66ABD'),
        (b'XXXXXXXabcdefg', b'3AA62DBBEFDB6763E0C510199CC66ABD'),
        (b'1234567XXXXXXX', b'0182BD0BD4444BF83AA62DBBEFDB6763'),
    ]

    def testKnownValues(self):
        """
        When lanman password are enabled in the configuration, they will
        return
        """
        cfg = config.loadConfig()
        cfg.set('samba', 'use-lmhash', 'yes')
        self.assertEqual(
            self.knownValues,
            [(p, smbpassword.lmhash(p)) for p, _ in self.knownValues]
        )

    def testLMHashConfiguationDisabled(self):
        """
        When the lanman passwords are disabled from the configuration, it will
        return the disabled password.
        """
        cfg = config.loadConfig()
        cfg.set('samba', 'use-lmhash', 'no')

        result = smbpassword.lmhash(b'any-pass')

        self.assertEqual(32 * b'X', result)



from twisted.internet import reactor
from twisted.web import server, resource

from ldaptor.protocols.ldap import ldapclient, ldapsyntax, ldapconnector, \
     distinguishedname
from ldaptor import ldapfilter

class LDAPConfig(object):
    def __init__(self,
                 baseDN,
                 serviceLocationOverrides=None):
        self.baseDN = distinguishedname.DistinguishedName(baseDN)
        self.serviceLocationOverrides = {}
        if serviceLocationOverrides is not None:
            for k,v in serviceLocationOverrides.items():
                dn = distinguishedname.DistinguishedName(k)
                self.serviceLocationOverrides[dn]=v

    def getBaseDN(self):
        return self.baseDN

    def getServiceLocationOverrides(self):
        return self.serviceLocationOverrides

class AddressBookResource(resource.Resource):
    def __init__(self, config):
        resource.Resource.__init__(self)
        self.config = config
        self.putChild('', self)

    def _search(self):
        c=ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d=c.connectAnonymously(self.config.getBaseDN(),
                               self.config.getServiceLocationOverrides())

        def _doSearch(proto):
            searchFilter = ldapfilter.parseFilter('(gn=j*)')
            baseEntry = ldapsyntax.LDAPEntry(client=proto,
                                             dn=self.config.getBaseDN())
            d=baseEntry.search(filterObject=searchFilter)
            return d

        d.addCallback(_doSearch)
        return d

    def _show(self, results, write):
        for item in results:
            write('<pre>')
            write(str(item))
            write('</pre>')

    def render(self, request):
        d = self._search()
        d.addCallback(self._show, request.write)
        d.addErrback(lambda e: request.write(str(e)))
        d.addBoth(lambda _: request.finish())
        return server.NOT_DONE_YET

def getSite(config):
    return server.Site(AddressBookResource(config))



"""
    Test cases for ldaptor.protocols.ldap.ldif module
"""

import base64

from twisted.trial import unittest

from ldaptor.protocols.ldap.ldif import attributeAsLDIF, asLDIF, manyAsLDIF


def encode(value):
    return b''.join(base64.encodestring(value).split(b'\n'))


class WireableObject(object):
    """
    Object with bytes representation as a constant toWire value
    """

    def toWire(self):
        return b'wire'


class AttributeAsLDIFTests(unittest.TestCase):
    """
    Converting pairs of attribute keys and values to LDIF.
    The result is a byte string with key and value
    separated by a semicolon and a space.
    In several special cases value is base64 encoded and separated
    from key by two semicolons and a space.
    """

    def test_byte_string(self):
        """Key and value are byte strings"""
        result = attributeAsLDIF(b'some key', b'some value')
        self.assertEqual(result, b'some key: some value\n')

    def test_unicode_string(self):
        """Key and value are unicode strings"""
        result = attributeAsLDIF(u'another key', u'another value')
        self.assertEqual(result, b'another key: another value\n')

    def test_wireable_object(self):
        """Value is an object with toWire method returning its bytes representation"""
        result = attributeAsLDIF('dn', WireableObject())
        self.assertEqual(result, b'dn: wire\n')

    def test_startswith_special_character(self):
        """
        Value is a string starting with one of the reserved characters.
        Returned value is base64 encoded.
        """
        for c in b'\0', b'\n', b'\r', b' ', b':', b'<':

            value = c + b'value'
            result = attributeAsLDIF(b'key', value)
            self.assertEqual(result, b'key:: %s\n' % encode(value))

    def test_endswith_special_character(self):
        """
        Value is a string ending with one of the reserved characters.
        Returned value is base64 encoded.
        """
        for c in b'\0', b'\n', b'\r', b' ':

            value = b'value' + c
            result = attributeAsLDIF(b'key', value)
            self.assertEqual(result, b'key:: %s\n' % encode(value))

    def test_contains_special_characters(self):
        """
        Value is a string with one of the reserved characters
        somewhere in its middle.
        Returned value is base64 encoded.
        """
        for c in b'\0', b'\n', b'\r':

            value = b'foo' + c + b'bar'
            result = attributeAsLDIF(b'key', value)
            self.assertEqual(result, b'key:: %s\n' % encode(value))

    def test_contains_nonprintable_characters(self):
        """
        Value is a string containing nonprintable characters.
        Returned value is base64 encoded.
        """
        result = attributeAsLDIF(b'key', b'val\xFFue')
        self.assertEqual(result, b'key:: %s\n' % encode(b'val\xFFue'))


class AsLDIFTests(unittest.TestCase):
    """
    Converting LDAP objects to LDIF.
    Object consists of DN and pairs of attribute keys and lists of values.
    The result is a number of lines in LDIF format for every key/value pair
    including DN.
    """

    def test_byte_string(self):
        """DN and attribute keys and values are byte strings"""
        attributes = [
            (b'key1', [b'value11', b'value12']),
            (b'key2', [b'value21', b'value22']),
        ]
        result = asLDIF(b'entry', attributes)
        self.assertEqual(result, b'''\
dn: entry
key1: value11
key1: value12
key2: value21
key2: value22

''')

    def test_unicode_string(self):
        """DN and attribute keys and values are unicode string"""
        attributes = [
            (u'key1', [u'value11', u'value12']),
            (u'key2', [u'value21', u'value22']),
        ]
        result = asLDIF(u'entry', attributes)
        self.assertEqual(result, b'''\
dn: entry
key1: value11
key1: value12
key2: value21
key2: value22

''')

    def test_wireable_object(self):
        """
        DN and attribute values are objects with
        toWire method returning bytes representation
        """
        attributes = [
            (b'key', [WireableObject()]),
        ]
        result = asLDIF(WireableObject(), attributes)
        self.assertEqual(result, b'dn: wire\nkey: wire\n\n')


class ManyAsLDIFTests(unittest.TestCase):
    """
    Converting multiple LDAP objects to LDIF.
    Every object consists of pair of DN and pairs of attribute keys and lists of values.
    The result is a number of blocks representing each object in LDIF format
    separated by an empty lines.
    The result contains a header with version number separated from other blocks
    by an empty line.
    """

    def test_multiple_objects(self):
        objects = [
            (b'object1', (
                (b'foo1', (b'value11', b'value12')),
                (b'foo2', (b'value21', b'value22')),
            )),
            (b'object2', (
                (b'bar1', (b'value31', b'value32')),
                (b'bar2', (b'value41', b'value42')),
            )),
        ]
        result = manyAsLDIF(objects)
        self.assertEqual(result, b'''\
version: 1

dn: object1
foo1: value11
foo1: value12
foo2: value21
foo2: value22

dn: object2
bar1: value31
bar1: value32
bar2: value41
bar2: value42

''')



# Copyright (C) 2001 Tommi Virtanen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import six

from ldaptor._encoder import to_bytes


def get(resultCode, errorMessage):
    """Get an instance of the correct exception for this resultCode."""
    return LDAPExceptionCollection.get_instance(resultCode, errorMessage)


class LDAPExceptionCollection(type):
    """
    Storage for the LDAP result codes and
    the corresponding classes.
    """

    collection = {}

    def __new__(mcs, name, bases, attributes):
        cls = type.__new__(mcs, name, bases, attributes)
        code = attributes.get('resultCode')
        if code is not None:
            assert isinstance(code, int)
            assert isinstance(attributes.get('name'), bytes)
            mcs.collection[code] = cls
        return cls

    @classmethod
    def get_instance(mcs, code, message):
        """Get an instance of the correct exception for this result code."""
        cls = mcs.collection.get(code)
        if cls is not None:
            return cls(message)
        return LDAPUnknownError(code, message)


class LDAPResult(six.with_metaclass(LDAPExceptionCollection)):
    resultCode = None
    name = None


class Success(LDAPResult):
    resultCode = 0
    name = b'success'

    def __init__(self, msg):
        pass


class LDAPException(Exception, LDAPResult):
    def __init__(self, message=None):
        Exception.__init__(self)
        self.message = message

    def __str__(self):
        message = self.toWire()
        return message if six.PY2 else message.decode('utf-8')

    def toWire(self):
        if self.message:
            return b'%s: %s' % (self.name, to_bytes(self.message))
        if self.name:
            return self.name
        return b'Unknown LDAP error %r' % self


class LDAPUnknownError(LDAPException):
    def __init__(self, resultCode, message=None):
        assert resultCode not in LDAPExceptionCollection.collection, \
            "resultCode %r must be unknown" % resultCode
        self.code = resultCode
        LDAPException.__init__(self, message)

    def toWire(self):
        codeName = b'unknownError(%d)' % self.code
        if self.message:
            return b'%s: %s' % (codeName, to_bytes(self.message))
        else:
            return codeName


class LDAPOperationsError(LDAPException):
    resultCode = 1
    name = b'operationsError'


class LDAPProtocolError(LDAPException):
    resultCode = 2
    name = b'protocolError'


class LDAPTimeLimitExceeded(LDAPException):
    resultCode = 3
    name = b'timeLimitExceeded'


class LDAPSizeLimitExceeded(LDAPException):
    resultCode = 4
    name = b'sizeLimitExceeded'


class LDAPCompareFalse(LDAPException):
    resultCode = 5
    name = b'compareFalse'


class LDAPCompareTrue(LDAPException):
    resultCode = 6
    name = b'compareTrue'


class LDAPAuthMethodNotSupported(LDAPException):
    resultCode = 7
    name = b'authMethodNotSupported'


class LDAPStrongAuthRequired(LDAPException):
    resultCode = 8
    name = b'strongAuthRequired'

# 9 reserved


class LDAPReferral(LDAPException):
    resultCode = 10
    name = b'referral'


class LDAPAdminLimitExceeded(LDAPException):
    resultCode = 11
    name = b'adminLimitExceeded'


class LDAPUnavailableCriticalExtension(LDAPException):
    resultCode = 12
    name = b'unavailableCriticalExtension'


class LDAPConfidentialityRequired(LDAPException):
    resultCode = 13
    name = b'confidentialityRequired'


class LDAPSaslBindInProgress(LDAPException):
    resultCode = 14
    name = b'saslBindInProgress'


class LDAPNoSuchAttribute(LDAPException):
    resultCode = 16
    name = b'noSuchAttribute'


class LDAPUndefinedAttributeType(LDAPException):
    resultCode = 17
    name = b'undefinedAttributeType'


class LDAPInappropriateMatching(LDAPException):
    resultCode = 18
    name = b'inappropriateMatching'


class LDAPConstraintViolation(LDAPException):
    resultCode = 19
    name = b'constraintViolation'


class LDAPAttributeOrValueExists(LDAPException):
    resultCode = 20
    name = b'attributeOrValueExists'


class LDAPInvalidAttributeSyntax(LDAPException):
    resultCode = 21
    name = b'invalidAttributeSyntax'

# 22-31 unused


class LDAPNoSuchObject(LDAPException):
    resultCode = 32
    name = b'noSuchObject'


class LDAPAliasProblem(LDAPException):
    resultCode = 33
    name = b'aliasProblem'


class LDAPInvalidDNSyntax(LDAPException):
    resultCode = 34
    name = b'invalidDNSyntax'

# 35 reserved for undefined isLeaf


class LDAPAliasDereferencingProblem(LDAPException):
    resultCode = 36
    name = b'aliasDereferencingProblem'

# 37-47 unused


class LDAPInappropriateAuthentication(LDAPException):
    resultCode = 48
    name = b'inappropriateAuthentication'


class LDAPInvalidCredentials(LDAPException):
    resultCode = 49
    name = b'invalidCredentials'


class LDAPInsufficientAccessRights(LDAPException):
    resultCode = 50
    name = b'insufficientAccessRights'


class LDAPBusy(LDAPException):
    resultCode = 51
    name = b'busy'


class LDAPUnavailable(LDAPException):
    resultCode = 52
    name = b'unavailable'


class LDAPUnwillingToPerform(LDAPException):
    resultCode = 53
    name = b'unwillingToPerform'


class LDAPLoopDetect(LDAPException):
    resultCode = 54
    name = b'loopDetect'

# 55-63 unused


class LDAPNamingViolation(LDAPException):
    resultCode = 64
    name = b'namingViolation'


class LDAPObjectClassViolation(LDAPException):
    resultCode = 65
    name = b'objectClassViolation'


class LDAPNotAllowedOnNonLeaf(LDAPException):
    resultCode = 66
    name = b'notAllowedOnNonLeaf'


class LDAPNotAllowedOnRDN(LDAPException):
    resultCode = 67
    name = b'notAllowedOnRDN'


class LDAPEntryAlreadyExists(LDAPException):
    resultCode = 68
    name = b'entryAlreadyExists'


class LDAPObjectClassModsProhibited(LDAPException):
    resultCode = 69
    name = b'objectClassModsProhibited'

# 70 reserved for CLDAP


class LDAPAffectsMultipleDSAs(LDAPException):
    resultCode = 71
    name = b'affectsMultipleDSAs'

# 72-79 unused


class LDAPOther(LDAPException):
    resultCode = 80
    name = b'other'

# 81-90 reserved for APIs


# Backwards compatibility
other = LDAPOther.resultCode
reverse = LDAPExceptionCollection.collection



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Testing a simple ldaptor ldap server
    Base on an example by Gaston TJEBBES aka "tonthon":
    http://tonthon.blogspot.com/2011/02/ldaptor-ldap-with-twisted-server-side.html
"""

import tempfile, sys

from twisted.application import service
from twisted.internet import reactor
from twisted.internet.protocol import ServerFactory
from twisted.python.components import registerAdapter
from twisted.python import log
from ldaptor.interfaces import IConnectedLDAPEntry
from ldaptor.protocols.ldap.ldapserver import LDAPServer
from ldaptor.ldiftree import LDIFTreeEntry

from schema import COUNTRY, COMPANY, PEOPLE, USERS


class Tree(object):

    def __init__(self):
        dirname = tempfile.mkdtemp('.ldap', 'test-server', '/tmp')
        self.db = LDIFTreeEntry(dirname)
        self.init_db()

    def init_db(self):
        """
            Add subtrees to the top entry
            top->country->company->people
        """
        country = self.db.addChild(COUNTRY[0], COUNTRY[1])
        company = country.addChild(COMPANY[0], COMPANY[1])
        people = company.addChild(PEOPLE[0], PEOPLE[1])
        for user in USERS:
            people.addChild(user[0], user[1])


class LDAPServerFactory(ServerFactory):
    """
        Our Factory is meant to persistently store the ldap tree
    """
    protocol = LDAPServer

    def __init__(self, root):
        self.root = root

    def buildProtocol(self, addr):
        proto = self.protocol()
        proto.debug = self.debug
        proto.factory = self
        return proto


if __name__ == '__main__':
    if len(sys.argv) == 2:
        port = int(sys.argv[1])
    else:
        port = 8080
    # First of all, to show logging info in stdout :
    log.startLogging(sys.stderr)
    # We initialize our tree
    tree = Tree()
    # When the ldap protocol handle the ldap tree,
    # it retrieves it from the factory adapting
    # the factory to the IConnectedLDAPEntry interface
    # So we need to register an adapter for our factory
    # to match the IConnectedLDAPEntry
    registerAdapter(
        lambda x: x.root,
        LDAPServerFactory,
        IConnectedLDAPEntry)
    # Run it !!
    factory = LDAPServerFactory(tree.db)
    factory.debug = True
    application = service.Application("ldaptor-server")
    myService = service.IServiceCollection(application)
    reactor.listenTCP(port, factory)
    reactor.run()



from twisted.python.failure import Failure
from twisted.internet import error
from ldaptor.protocols.ldap import ldifprotocol
from ldaptor import delta, entry

WAIT_FOR_CHANGETYPE = b'WAIT_FOR_CHANGETYPE'
WAIT_FOR_MOD_SPEC = b'WAIT_FOR_MOD_SPEC'
IN_MOD_SPEC = b'IN_MOD_SPEC'
IN_ADD_ENTRY = b'IN_ADD_ENTRY'
IN_DELETE = b'IN_DELETE'

class LDIFDeltaMissingChangeTypeError(ldifprotocol.LDIFParseError):
    """LDIF delta entry has no changetype."""

class LDIFDeltaUnknownChangeTypeError(ldifprotocol.LDIFParseError):
    """LDIF delta entry has an unknown changetype."""


class LDIFDeltaUnknownModificationError(ldifprotocol.LDIFParseError):
    """LDIF delta modification has unknown mod-spec."""



class LDIFDeltaModificationMissingEndDashError(ldifprotocol.LDIFParseError):
    """LDIF delta modification has no ending dash."""



class LDIFDeltaModificationDifferentAttributeTypeError(ldifprotocol.LDIFParseError):
    """The attribute type for the change is not the as in the mod-spec header line."""



class LDIFDeltaAddMissingAttributesError(ldifprotocol.LDIFParseError):
    """Add operation needs to have at least one attribute type and value."""



class LDIFDeltaDeleteHasJunkAfterChangeTypeError(ldifprotocol.LDIFParseError):
    """Delete operation takes no attribute types or values."""



class LDIFDelta(ldifprotocol.LDIF):
    def state_WAIT_FOR_DN(self, line):
        super(LDIFDelta, self).state_WAIT_FOR_DN(line)
        if self.mode == ldifprotocol.IN_ENTRY:
            self.mode = WAIT_FOR_CHANGETYPE

    def state_WAIT_FOR_CHANGETYPE(self, line):
        assert self.dn is not None, 'self.dn must be set when in entry'
        assert self.data is not None, 'self.data must be set when in entry'

        if line == b'':
            raise LDIFDeltaMissingChangeTypeError(self.dn)

        key, val = self._parseLine(line)

        if key != b'changetype':
            raise LDIFDeltaMissingChangeTypeError(self.dn, key, val)

        if val == b'modify':
            self.modifications = []
            self.mode = WAIT_FOR_MOD_SPEC
        elif val == b'add':
            self.mode = IN_ADD_ENTRY
        elif val == b'delete':
            self.mode = IN_DELETE
        elif val == b'modrdn' or val == b'moddn':
            raise NotImplementedError()  # TODO
        else:
            raise LDIFDeltaUnknownChangeTypeError()

    MOD_SPEC_TO_DELTA = {
        b'add': delta.Add,
        b'delete': delta.Delete,
        b'replace': delta.Replace,
        }

    def state_WAIT_FOR_MOD_SPEC(self, line):
        if line == b'':
            # end of entry
            self.mode = ldifprotocol.WAIT_FOR_DN
            m = delta.ModifyOp(dn=self.dn,
                               modifications=self.modifications)
            self.dn = None
            self.data = None
            self.modifications = None
            self.gotEntry(m)
            return

        key, val = self._parseLine(line)

        if key not in self.MOD_SPEC_TO_DELTA:
            raise LDIFDeltaUnknownModificationError(self.dn, key)

        self.mod_spec = key
        self.mod_spec_attr = val
        self.mod_spec_data = []
        self.mode = IN_MOD_SPEC

    def state_IN_MOD_SPEC(self, line):
        if line == b'':
            raise LDIFDeltaModificationMissingEndDashError()

        if line == b'-':
            mod = self.MOD_SPEC_TO_DELTA[self.mod_spec]
            de = mod(self.mod_spec_attr, self.mod_spec_data)
            self.modifications.append(de)
            del self.mod_spec
            del self.mod_spec_attr
            del self.mod_spec_data
            self.mode = WAIT_FOR_MOD_SPEC
            return

        key, val = self._parseLine(line)

        if key != self.mod_spec_attr:
            raise LDIFDeltaModificationDifferentAttributeTypeError(
                key, self.mod_spec_attr)

        self.mod_spec_data.append(val)

    def state_IN_ADD_ENTRY(self, line):
        assert self.dn is not None, 'self.dn must be set when in entry'
        assert self.data is not None, 'self.data must be set when in entry'

        if line == b'':
            # end of entry
            if not self.data:
                raise LDIFDeltaAddMissingAttributesError(self.dn)
            self.mode = ldifprotocol.WAIT_FOR_DN
            o = delta.AddOp(entry.BaseLDAPEntry(dn=self.dn,
                                                attributes=self.data))
            self.dn = None
            self.data = None
            self.gotEntry(o)
            return

        key, val = self._parseLine(line)

        if not key in self.data:
            self.data[key] = []

        self.data[key].append(val)

    def state_IN_DELETE(self, line):
        assert self.dn is not None, 'self.dn must be set when in entry'

        if line == b'':
            # end of entry
            self.mode = ldifprotocol.WAIT_FOR_DN
            o = delta.DeleteOp(dn=self.dn)
            self.dn = None
            self.data = None
            self.gotEntry(o)
            return

        raise LDIFDeltaDeleteHasJunkAfterChangeTypeError(
                self.dn, line)

def fromLDIFFile(f):
    """Read LDIF data from a file."""

    p = LDIFDelta()
    l = []
    p.gotEntry = l.append
    while 1:
        data = f.read()
        if not data:
            break
        p.dataReceived(data)
    p.connectionLost(Failure(error.ConnectionDone()))

    return l



"""LDAP protocol server, which acts as a proxy which
   forwards the requests to multiple LDAP servers and
   merges the results.
   Only Bind and Search requests are supported.
"""

from twisted.internet import reactor, defer
from six.moves.queue import Queue

from ldaptor.protocols.ldap import ldapclient, ldapconnector
from ldaptor.protocols.ldap import ldapserver
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols import pureldap
from ldaptor.config import LDAPConfig


class MergedLDAPServer(ldapserver.BaseLDAPServer):
    protocol = ldapclient.LDAPClient

    def __init__(self, configs, use_tls):
        ldapserver.BaseLDAPServer.__init__(self)
        self.clients = []
        self.configs = configs
        self.use_tls = use_tls
        self.all_connected = False
        self.merge_map = {}
        self.waitingConnect = []
        self.unbound = False

    def _whenConnected(self, fn, *a, **kw):
        if not self.all_connected:
            d = defer.Deferred()
            self.waitingConnect.append((d, fn, a, kw))
            return d
        else:
            return defer.maybeDeferred(fn, *a, **kw)

    def _failConnection(self, reason):
        self.transport.loseConnection()
        raise ldaperrors.LDAPOther("Cannot connect to server.{}".format(reason))

    def _cbConnectionMade(self, proto):
        self.clients.append(proto)

        if len(self.clients) == len(self.configs):
            self.all_connected = True

        # Only call once when all clients are connected.
        if self.all_connected:
            while self.waitingConnect:
                d, fn, a, kw = self.waitingConnect.pop(0)
                d2 = defer.maybeDeferred(fn, *a, **kw)
                d2.chainDeferred(d)

    def _clientQueue(self, request, controls, reply):
        # Controls are ignored.
        for c in self.clients:
            if request.needs_answer:
                d = c.send_multiResponse(request, self._gotResponse, reply)
                d.addErrback(defer.logError)
            else:
                c.send_noResponse(request)

    def queue(self, id, op):
        if isinstance(op, (pureldap.LDAPSearchResultDone,
                           pureldap.LDAPBindResponse)):
            if id not in self.merge_map:
                self.merge_map[id] = Queue(len(self.clients))
                self.merge_map[id].put(op)
            else:
                self.merge_map[id].put(op)

            if self.merge_map[id].full():
                # Send success, if at least one success.
                for i in range(len(self.clients)):
                    r = self.merge_map[id].get()
                    if r.resultCode == ldaperrors.Success.resultCode:
                        op = r
                del self.merge_map[id]
                ldapserver.BaseLDAPServer.queue(self, id, op)
        else:
            ldapserver.BaseLDAPServer.queue(self, id, op)

    def connectionMade(self):
        clientCreator = ldapconnector.LDAPClientCreator(
            reactor, self.protocol)
        for (c, tls) in zip(self.configs, self.use_tls):
            d = clientCreator.connect(dn='',
                                      overrides=c.getServiceLocationOverrides())
            if tls:
                d.addCallback(lambda x: x.startTLS())
            d.addCallback(self._cbConnectionMade)
            d.addErrback(self._failConnection)

        ldapserver.BaseLDAPServer.connectionMade(self)

    def connectionLost(self, reason):
        for c in self.clients:
            assert c is not None
            if c.connected:
                if not self.unbound:
                    c.unbind()
                else:
                    c.transport.loseConnection()

        self.clients = []
        self.unbound = True
        ldapserver.BaseLDAPServer.connectionLost(self, reason)

    def _gotResponse(self, response, reply):
        reply(response)

        # TODO this is ugly
        return isinstance(response, (
            pureldap.LDAPSearchResultDone,
            pureldap.LDAPBindResponse,
            ))

    def _handleUnknown(self, request, controls, reply):
        self._whenConnected(self._clientQueue, request, controls, reply)
        return None

    def handleUnknown(self, request, controls, reply):
        d = defer.succeed(request)
        d.addCallback(self._handleUnknown, controls, reply)
        return d

    def handle_LDAPBindRequest(self, request, controls, reply):
        return self.handleUnknown(request, controls, reply)

    def handle_LDAPSearchRequest(self, request, controls, reply):
        return self.handleUnknown(request, controls, reply)

    def handle_LDAPUnbindRequest(self, request, controls, reply):
        self.unbound = True
        self.handleUnknown(request, controls, reply)

    fail_LDAPDelRequest = pureldap.LDAPDelResponse

    def handle_LDAPDelRequest(self, request, controls, reply):
         raise ldaperrors.LDAPUnwillingToPerform()

    fail_LDAPAddRequest = pureldap.LDAPAddResponse

    def handle_LDAPAddRequest(self, request, controls, reply):
        raise ldaperrors.LDAPUnwillingToPerform()

    fail_LDAPModifyDNRequest = pureldap.LDAPModifyDNResponse

    def handle_LDAPModifyDNRequest(self, request, controls, reply):
         raise ldaperrors.LDAPUnwillingToPerform()

    fail_LDAPModifyRequest = pureldap.LDAPModifyResponse

    def handle_LDAPModifyRequest(self, request, controls, reply):
        raise ldaperrors.LDAPUnwillingToPerform()

    fail_LDAPExtendedRequest = pureldap.LDAPExtendedResponse

    def handle_LDAPExtendedRequest(self, request, controls, reply):
         raise ldaperrors.LDAPUnwillingToPerform()

if __name__ == '__main__':
    from twisted.internet import protocol
    from twisted.python import log
    import sys
    log.startLogging(sys.stderr)

    configs = [LDAPConfig(serviceLocationOverrides={"": ('localhost', 38942)}),
               LDAPConfig(serviceLocationOverrides={"": ('localhost', 8080)})]
    use_tls = [True, False]
    factory = protocol.ServerFactory()
    factory.protocol = lambda: MergedLDAPServer(configs, use_tls)
    reactor.listenTCP(10389, factory)
    reactor.run()



"""
Test cases for ldaptor.protocols.ldap.ldapsyntax module.
"""

from twisted.trial import unittest
from ldaptor import config, testutil, delta
from ldaptor.protocols.ldap import ldapclient, ldapsyntax, ldaperrors
from ldaptor.protocols import pureldap, pureber
from twisted.internet import defer
from twisted.internet import error
from twisted.python import failure
from ldaptor.testutil import LDAPClientTestDriver

class LDAPEntryTests(unittest.TestCase):
    """
    Unit tests for LDAPEntry.
    """

    def testCreation(self):
        """Creating an LDAP object should succeed."""
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o['objectClass'], ['a', 'b'])
        self.failUnlessEqual(o['aValue'], ['a'])
        self.failUnlessEqual(o['bValue'], ['b'])
        client.assertNothingSent()

    def testKeys(self):
        """Iterating over the keys of an LDAP object gives expected results."""
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        seen={}
        for k in o.keys():
            assert k not in seen
            seen[k]=1
        assert seen == {'objectClass': 1,
                        'aValue': 1,
                        'bValue': 1,
                        }

    def testItems(self):
        """Iterating over the items of an LDAP object gives expected results."""
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        seen={}
        for k,vs in o.items():
            assert k not in seen
            seen[k]=vs
        assert seen == {'objectClass': ['a', 'b'],
                        'aValue': ['a'],
                        'bValue': ['b'],
                        }

    def testIn(self):
        """Key in object gives expected results."""
        client=LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        assert 'objectClass' in o
        assert 'aValue' in o
        assert 'bValue' in o
        assert 'foo' not in o
        assert '' not in o
        assert None not in o

        assert 'a' in o['objectClass']
        assert 'b' in o['objectClass']
        assert 'foo' not in o['objectClass']
        assert '' not in o['objectClass']
        assert None not in o['objectClass']

        assert 'a' in o['aValue']
        assert 'foo' not in o['aValue']
        assert '' not in o['aValue']
        assert None not in o['aValue']

    def testInequalityOtherObject(self):
        """
        It is not equal with non LDAPEntry objects.
        """
        client = LDAPClientTestDriver()
        sut = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            )

        self.assertNotEqual('dc=example,dc=com', sut)

    def testInequalityDN(self):
        """
        Entries with different DN are not equal.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=org',
            )

        self.assertNotEqual(first, second)

    def testInequalityAttributes(self):
        """
        Entries with same DN but different attributes are not equal.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={'attr_key1': ['some-value']},
            )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={'attr_key2': ['some-value']},
            )

        self.assertNotEqual(first, second)

    def testInequalityValues(self):
        """
        Entries with same DN same attributes, but different
        values for attributes are not equal.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={'attr_key1': ['some-value']},
            )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={'attr_key1': ['other-value']},
            )

        self.assertNotEqual(first, second)

    def testEquality(self):
        """
        Entries with same DN, same attributes, and same values for
        attributes equal, regardless of the order of the attributes.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={
                'attr_key1': ['some-value'],
                'attr_key2': ['second-value'],
                },
            )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={
                'attr_key2': ['second-value'],
                'attr_key1': ['some-value'],
                },
            )

        self.assertEqual(first, second)

    def testHashEqual(self):
        """
        Entries which are equal have the same hash.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            )

        self.assertEqual(first, second)
        self.assertEqual(hash(first), hash(second))

    def testHashNotEqual(self):
        """
        Entries which are not equal have different hash values.
        """
        client = LDAPClientTestDriver()
        first = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            )
        second = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=org',
            )

        self.assertNotEqual(first, second)
        self.assertNotEqual(hash(first), hash(second))


class LDAPSyntaxAttributes(unittest.TestCase):
    def testAttributeSetting(self):
        client=LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        o['aValue']=['foo', 'bar']
        self.failUnlessEqual(o['aValue'], ['foo', 'bar'])
        o['aValue']=['quux']
        self.failUnlessEqual(o['aValue'], ['quux'])
        self.failUnlessEqual(o['bValue'], ['b'])
        o['cValue']=['thud']
        self.failUnlessEqual(o['aValue'], ['quux'])
        self.failUnlessEqual(o['bValue'], ['b'])
        self.failUnlessEqual(o['cValue'], ['thud'])

    def testAttributeDelete(self):
        client=LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        o['aValue']=['quux']
        del o['aValue']
        del o['bValue']
        self.failIf('aValue' in o)
        self.failIf('bValue' in o)

    def testAttributeAdd(self):
        client=LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        o['aValue'].add('foo')
        self.failUnlessEqual(o['aValue'], ['a', 'foo'])

    def testAttributeItemDelete(self):
        client=LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a', 'b', 'c'],
            'bValue': ['b'],
            })
        o['aValue'].remove('b')
        self.failUnlessEqual(o['aValue'], ['a', 'c'])

    def testUndo(self):
        """Undo should forget the modifications."""
        client=LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            'cValue': ['c'],
            })
        o['aValue']=['foo', 'bar']
        o['aValue']=['quux']
        del o['cValue']
        o.undo()
        self.failUnlessEqual(o['aValue'], ['a'])
        self.failUnlessEqual(o['bValue'], ['b'])
        self.failUnlessEqual(o['cValue'], ['c'])

    def testUndoJournaling(self):
        """Journaling should still work after undo."""
        client=LDAPClientTestDriver(
            [ pureldap.LDAPModifyResponse(resultCode=0,
                                          matchedDN='',
                                          errorMessage=''),
            ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            'cValue': ['c'],
            })
        o['aValue']=['foo', 'bar']
        o['aValue']=['quux']
        del o['cValue']
        o.undo()
        o['aValue'].update(['newValue', 'anotherNewValue'])
        d=o.commit()
        def cb(dummy):
            self.failUnlessEqual(o['aValue'], ['a', 'newValue', 'anotherNewValue'])
            self.failUnlessEqual(o['bValue'], ['b'])
            self.failUnlessEqual(o['cValue'], ['c'])
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Add('aValue', ['newValue', 'anotherNewValue']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testUndoAfterCommit(self):
        """Undo should not undo things that have been commited."""

        client=LDAPClientTestDriver(
            [ pureldap.LDAPModifyResponse(resultCode=0,
                                          matchedDN='',
                                          errorMessage=''),
            ])
        o=ldapsyntax.LDAPEntry(
            client=client,
            dn='cn=foo,dc=example,dc=com',
            attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            'cValue': ['c'],
            })
        o['aValue']=['foo', 'bar']
        o['bValue']=['quux']
        del o['cValue']

        d=o.commit()
        def cb(dummy):
            o.undo()
            self.failUnlessEqual(o['aValue'], ['foo', 'bar'])
            self.failUnlessEqual(o['bValue'], ['quux'])
            self.failIf('cValue' in o)
        d.addCallback(cb)
        return d

class LDAPSyntaxAttributesModificationOnWire(unittest.TestCase):
    def testAdd(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyResponse(resultCode=0,
                                            matchedDN='',
                                            errorMessage=''),
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            })
        o['aValue'].update(['newValue', 'anotherNewValue'])

        d=o.commit()
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Add('aValue', ['newValue', 'anotherNewValue']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testAddSeparate(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyResponse(resultCode=0,
                                            matchedDN='',
                                            errorMessage=''),
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            })
        o['aValue'].add('newValue')
        o['aValue'].add('anotherNewValue')

        d=o.commit()
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Add('aValue', ['newValue']),
                delta.Add('aValue', ['anotherNewValue']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testDeleteAttribute(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyResponse(resultCode=0,
                                            matchedDN='',
                                            errorMessage='')
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            })
        o['aValue'].remove('a')

        d=o.commit()
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Delete('aValue', ['a']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testDeleteAllAttribute(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyResponse(resultCode=0,
                                            matchedDN='',
                                            errorMessage=''),
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a1', 'a2'],
            'bValue': ['b1', 'b2'],
            })
        del o['aValue']
        o['bValue'].clear()

        d=o.commit()
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Delete('aValue'),
                delta.Delete('bValue'),
                ]).asLDAP())
        d.addCallback(cb)
        return d


    def testReplaceAttributes(self):
        """Modify & commit should write the right data to the server."""

        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyResponse(resultCode=0,
                                            matchedDN='',
                                            errorMessage='')
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            })
        o['aValue']=['foo', 'bar']

        d=o.commit()
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('aValue', ['foo', 'bar']),
                ]).asLDAP())
        d.addCallback(cb)
        return d


class LDAPSyntaxSearch(unittest.TestCase):
    timeout = 3

    def _test_search(self, return_controls=False):
        """
        Create a test search.
        Return Deferred with no handler.
        """
        client=LDAPClientTestDriver([
            pureldap.LDAPSearchResultEntry(
                objectName='cn=foo,dc=example,dc=com',
                attributes=(
                    ('foo', ['a']),
                    ('bar', ['b', 'c']),
                ),
            ),
            pureldap.LDAPSearchResultEntry(
                objectName='cn=bar,dc=example,dc=com',
                attributes=(
                    ('foo', ['a']),
                    ('bar', ['d', 'e']),
                ),
            ),
            pureldap.LDAPSearchResultDone(
                resultCode=0,
                matchedDN='',
                errorMessage='')
            ])
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={
                'objectClass': ['organizationalUnit'],
            }
        )
        d = o.search(
            filterText='(foo=a)',
            attributes=['foo', 'bar'],
            return_controls=return_controls)

        def cb(val):
            if return_controls:
                val = val[0]
            client.assertSent(
                pureldap.LDAPSearchRequest(
                    baseObject='dc=example,dc=com',
                    scope=pureldap.LDAP_SCOPE_wholeSubtree,
                    derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                    sizeLimit=0,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(value='foo'),
                        assertionValue=pureldap.LDAPAssertionValue(value='a')
                    ),
                    attributes=['foo', 'bar']
                )
            )
            self.failUnlessEqual(len(val), 2)
            self.failUnlessEqual(
                val[0],
                ldapsyntax.LDAPEntry(
                    client=client,
                    dn='cn=foo,dc=example,dc=com',
                    attributes={
                        b'foo': [b'a'],
                        b'bar': [b'b', b'c'],
                    }
                )
            )
            self.failUnlessEqual(
                val[1],
                ldapsyntax.LDAPEntry(
                    client=client,
                    dn='cn=bar,dc=example,dc=com',
                    attributes={
                        b'foo': [b'a'],
                        b'bar': [b'd', b'e'],
                    }
                )
            )

        d.addCallback(cb)
        return d

    def testSearch(self):
        """Test searches."""
        return self._test_search()

    def test_search_not_connected(self):
        client = ldapclient.LDAPClient()
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={
                'objectClass': ['organizationalUnit'],
            }
        )
        d = o.search(
            filterText='(foo=a)',
            attributes=['foo', 'bar'])

        def cb_(thing):
            self.assertEqual(
                ldapclient.LDAPClientConnectionLostException,
                type(thing.value))

        d.addErrback(cb_)
        return d
        

    def test_search_controls_returned(self):
        return self._test_search(return_controls=True)

    def test_search_size_limit_exceeded(self):
        client=LDAPClientTestDriver([
            pureldap.LDAPSearchResultEntry(
                objectName='cn=foo,dc=example,dc=com',
                attributes=(
                    ('foo', ['a']),
                    ('bar', ['b', 'c']),
                ),
            ),
            pureldap.LDAPSearchResultDone(
                resultCode=ldaperrors.LDAPSizeLimitExceeded.resultCode,
                matchedDN='',
                errorMessage='Size limit exceeded.')
            ])
        o = ldapsyntax.LDAPEntry(
            client=client,
            dn='dc=example,dc=com',
            attributes={
                'objectClass': ['organizationalUnit'],
            }
        )
        d = o.search(
            filterText='(foo=a)',
            attributes=['foo', 'bar'],
            sizeLimit=1,
            return_controls=False)

        def cb_(thing):
            self.failUnlessEqual(len(thing), 1)

        d.addCallback(cb_)
        d.addErrback(cb_)
        return d

    def testSearch_defaultAttributes(self):
        """Search without explicit list of attributes returns all attributes."""

        client=LDAPClientTestDriver([
            pureldap.LDAPSearchResultEntry(
            objectName='cn=foo,dc=example,dc=com',
            attributes=(('foo', ['a']),
                        ('bar', ['b', 'c']),
                        ),
            ),
            pureldap.LDAPSearchResultEntry(
            objectName='cn=bar,dc=example,dc=com',
            attributes=(('foo', ['a']),
                        ('bar', ['d', 'e']),
                        ),
            ),
            pureldap.LDAPSearchResultDone(
            resultCode=0,
            matchedDN='',
            errorMessage='')
            ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='dc=example,dc=com',
                               attributes={
            'objectClass': ['organizationalUnit'],
            })

        d=o.search(filterText='(foo=a)')
        def cb(val):
            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_wholeSubtree,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value='foo'),
                assertionValue=pureldap.LDAPAssertionValue(value='a')),
                attributes=[]))
            self.failUnlessEqual(len(val), 2)

            self.failUnlessEqual(val[0],
                                 ldapsyntax.LDAPEntry(
                client=client,
                dn='cn=foo,dc=example,dc=com',
                attributes={
                b'foo': [b'a'],
                b'bar': [b'b', b'c'],
                }))
            self.failUnless(val[0].complete)

            self.failUnlessEqual(val[1],
                                 ldapsyntax.LDAPEntry(
                client=client,
                dn='cn=bar,dc=example,dc=com',
                attributes={
                b'foo': [b'a'],
                b'bar': [b'd', b'e'],
                }))
            self.failUnless(val[1].complete)
        d.addCallback(cb)
        return d

    def testSearch_noAttributes(self):
        """Search with attributes=None returns no attributes."""

        client=LDAPClientTestDriver([
            pureldap.LDAPSearchResultEntry('cn=foo,dc=example,dc=com',
                                           attributes=()),
            pureldap.LDAPSearchResultEntry('cn=bar,dc=example,dc=com',
                                           attributes=()),
            pureldap.LDAPSearchResultDone(
            resultCode=0,
            matchedDN='',
            errorMessage='')
            ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='dc=example,dc=com',
                               attributes={
            'objectClass': ['organizationalUnit'],
            })

        d=o.search(filterText='(foo=a)',
                   attributes=None)
        def cb(val):
            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_wholeSubtree,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value='foo'),
                assertionValue=pureldap.LDAPAssertionValue(value='a')),
                attributes=['1.1']))
            self.failUnlessEqual(len(val), 2)

            self.failUnlessEqual(val[0],
                                 ldapsyntax.LDAPEntry(
                client=client,
                dn='cn=foo,dc=example,dc=com'))
            self.failIf(val[0].complete)

            self.failUnlessEqual(val[1],
                                 ldapsyntax.LDAPEntry(
                client=client,
                dn='cn=bar,dc=example,dc=com'))
            self.failIf(val[1].complete)
        d.addCallback(cb)
        return d

    def testSearch_ImmediateProcessing(self):
        """Test searches with the immediate processing feature."""

        client=LDAPClientTestDriver([
            pureldap.LDAPSearchResultEntry(
            objectName='cn=foo,dc=example,dc=com',
            attributes=(('bar', ['b', 'c']),
                        ),
            ),

            pureldap.LDAPSearchResultEntry(
            objectName='cn=bar,dc=example,dc=com',
            attributes=(('bar', ['b', 'c']),
                        ),
            ),

            pureldap.LDAPSearchResultDone(
            resultCode=0,
            matchedDN='',
            errorMessage='')
            ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='dc=example,dc=com',
                               attributes={
            'objectClass': ['organizationalUnit'],
            })

        seen=[]
        def process(o):
            seen.append(o)

        d=o.search(filterText='(foo=a)',
                   attributes=['bar'],
                   callback=process)
        def cb(val):
            self.assertEqual(val, None)

            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_wholeSubtree,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value='foo'),
                assertionValue=pureldap.LDAPAssertionValue(value='a')),
                attributes=['bar']))

            self.failUnlessEqual(seen,
                                 [
                ldapsyntax.LDAPEntry(
                client=client,
                dn='cn=foo,dc=example,dc=com',
                attributes={
                b'bar': [b'b', b'c'],
                }),
                ldapsyntax.LDAPEntry(
                client=client,
                dn='cn=bar,dc=example,dc=com',
                attributes={
                b'bar': [b'b', b'c'],
                })])
        d.addCallback(cb)
        return d

    def testSearch_fail(self):
        client=LDAPClientTestDriver([
            pureldap.LDAPSearchResultDone(
            resultCode=ldaperrors.LDAPBusy.resultCode,
            matchedDN='',
            errorMessage='Go away')
            ])

        o=ldapsyntax.LDAPEntry(client=client, dn='dc=example,dc=com')
        d=o.search(filterText='(foo=a)')
        def eb(fail):
            fail.trap(ldaperrors.LDAPBusy)
            self.assertEqual(fail.value.message, 'Go away')

            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_wholeSubtree,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(value='foo'),
                assertionValue=pureldap.LDAPAssertionValue(value='a')),
                ))
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testSearch_err(self):
        client=LDAPClientTestDriver([
                failure.Failure(error.ConnectionLost())
                ])
        o = ldapsyntax.LDAPEntry(client=client, dn='dc=example,dc=com')
        d = o.search(filterText='(foo=a)')
        def eb(fail):
            fail.trap(error.ConnectionLost)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

class LDAPSyntaxDNs(unittest.TestCase):
    def testDNKeyExistenceSuccess(self):
        client = LDAPClientTestDriver()
        ldapsyntax.LDAPEntry(client=client,
                             dn='cn=foo,dc=example,dc=com',
                             attributes={
            'cn': ['foo'],
            })


class LDAPSyntaxLDIF(unittest.TestCase):
    def testLDIFConversion(self):
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a', 'b'],
            'bValue': ['c'],
            })
        self.failUnlessEqual(o.toWire(),
                             b'''dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

''')


class LDAPSyntaxDelete(unittest.TestCase):
    def testDeleteInvalidates(self):
        """Deleting an LDAPEntry invalidates it."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPDelResponse(resultCode=0,
                                      matchedDN='',
                                      errorMessage=''),
             ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a'],
            })
        d=o.delete()
        def cb(dummy):
            self.failUnlessRaises(
                ldapsyntax.ObjectDeletedError,
                o.search,
                filterText='(foo=a)')
            self.failUnlessRaises(
                ldapsyntax.ObjectDeletedError,
                o.get,
                'objectClass')
        d.addCallback(cb)
        return d

    def testDeleteOnWire(self):
        """LDAPEntry.delete should write the right data to the server."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPDelResponse(resultCode=0,
                                      matchedDN='',
                                      errorMessage=''),
             ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a'],
            })
        d=o.delete()
        def cb(dummy):
            client.assertSent(pureldap.LDAPDelRequest(
                entry='cn=foo,dc=example,dc=com',
                ))
        d.addCallback(cb)
        return d

    def testErrorHandling(self):
        """LDAPEntry.delete should pass LDAP errors to it's deferred."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPDelResponse(resultCode=ldaperrors.LDAPBusy.resultCode,
                                      matchedDN='',
                                      errorMessage='Go away'),
             ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a'],
            })
        d=o.delete()
        def eb(fail):
            fail.trap(ldaperrors.LDAPBusy)
            self.assertEqual(fail.value.message, 'Go away')

            client.assertSent(pureldap.LDAPDelRequest(
                entry='cn=foo,dc=example,dc=com',
                ))
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testErrorHandling_extended(self):
        """LDAPEntry.delete should pass even non-LDAPDelResponse errors to it's deferred."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=ldaperrors.LDAPProtocolError.resultCode,
                                                 responseName='1.3.6.1.4.1.1466.20036',
                                                 errorMessage='Unknown request')
             ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a'],
            })
        d=o.delete()
        def eb(fail):
            fail.trap(ldaperrors.LDAPProtocolError)
            self.assertEqual(fail.value.message, 'Unknown request')

            client.assertSent(pureldap.LDAPDelRequest(
                entry='cn=foo,dc=example,dc=com',
                ))
        d.addCallbacks(testutil.mustRaise, eb)
        return d

class LDAPSyntaxAddChild(unittest.TestCase):
    def testAddChildOnWire(self):
        """LDAPEntry.addChild should write the right data to the server."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPAddResponse(resultCode=0,
                                      matchedDN='',
                                      errorMessage=''),
             ])
        sut = ldapsyntax.LDAPEntry(
            client=client,
            dn='ou=things,dc=example,dc=com',
            attributes={
                'objectClass': ['organizationalUnit'],
                'ou': ['things'],
                },
            )
        d = sut.addChild(
            rdn='givenName=Firstname+surname=Lastname',
            attributes={
                'objectClass': ['person', 'otherStuff'],
                'givenName': ['Firstname'],
                'surname': ['Lastname'],
                },
            )
        self.successResultOf(d)

        client.assertSent(pureldap.LDAPAddRequest(
            entry='givenName=Firstname+surname=Lastname,ou=things,dc=example,dc=com',
            attributes=[
                (
                    pureldap.LDAPAttributeDescription('objectClass'),
                    pureber.BERSet([
                        pureldap.LDAPAttributeValue('person'),
                        pureldap.LDAPAttributeValue('otherStuff'),
                        ]),
                    ),
                (
                    pureldap.LDAPAttributeDescription('givenName'),
                    pureber.BERSet([pureldap.LDAPAttributeValue('Firstname')])),
                    (
                        pureldap.LDAPAttributeDescription('surname'),
                        pureber.BERSet([pureldap.LDAPAttributeValue('Lastname')],
                            ),
                        ),
            ],
            ))


class LDAPSyntaxContainingNamingContext(unittest.SynchronousTestCase):
    def setUp(self):
        attributes = [
            (
                'namingContexts',
                (
                    'dc=foo,dc=example',
                    'dc=example,dc=com',
                    'dc=bar,dc=example',
                )
            )
        ]
        self.client = LDAPClientTestDriver([
            pureldap.LDAPSearchResultEntry(objectName='', attributes=attributes),
            pureldap.LDAPSearchResultDone(resultCode=0, matchedDN='', errorMessage='')
        ])

    def testNamingContext(self):
        """LDAPEntry.namingContext returns the naming context that contains this object (via a Deferred)."""
        o = ldapsyntax.LDAPEntry(
            client=self.client,
            dn='cn=foo,ou=bar,dc=example,dc=com',
            attributes={'objectClass': ['a']}
        )
        d = o.namingContext()
        def cb(p):
            assert isinstance(p, ldapsyntax.LDAPEntry)
            assert p.client == o.client
            assert p.dn.getText() == u'dc=example,dc=com'

            self.client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='',
                scope=pureldap.LDAP_SCOPE_baseObject,
                filter=pureldap.LDAPFilter_present('objectClass'),
                attributes=['namingContexts'],
                ))
        d.addCallback(cb)
        return d

    def testNoContainingNamingContext(self):
        """LDAPEntry.namingContext raises exception if there are no naming contexts with it"""
        o = ldapsyntax.LDAPEntry(
            client=self.client,
            dn='cn=foo,dc=foo,dc=com',
            attributes={'objectClass': ['a']}
        )
        fail = self.failureResultOf(o.namingContext())
        self.assertIsInstance(fail.value, ldapsyntax.NoContainingNamingContext)


class LDAPSyntaxPasswords(unittest.TestCase):
    def setUp(self):
        cfg = config.loadConfig()
        cfg.set('samba', 'use-lmhash', 'no')

    def testPasswordSetting_ExtendedOperation(self):
        """LDAPEntry.setPassword_ExtendedOperation(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d=o.setPassword_ExtendedOperation(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              )
        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_sambaAccount(self):
        """LDAPEntry.setPassword_Samba(newPasswd=...,
        style='sambaAccount') changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d=o.setPassword_Samba(newPasswd=b'new', style='sambaAccount')
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('ntPassword',
                              ['89963F5042E5041A59C249282387A622']),
                delta.Replace('lmPassword',
                              ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_sambaSamAccount(self):
        """LDAPEntry.setPassword_Samba(newPasswd=..., style='sambaSamAccount') changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d=o.setPassword_Samba(newPasswd=b'new', style='sambaSamAccount')
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('sambaNTPassword',
                              ['89963F5042E5041A59C249282387A622']),
                delta.Replace('sambaLMPassword',
                              ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_defaultStyle(self):
        """LDAPEntry.setPassword_Samba(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d=o.setPassword_Samba(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('sambaNTPassword',
                              ['89963F5042E5041A59C249282387A622']),
                            delta.Replace('sambaLMPassword',
                                          ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d

    def testPasswordSetting_Samba_badStyle(self):
        """LDAPEntry.setPassword_Samba(..., style='foo') fails."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d=defer.maybeDeferred(o.setPassword_Samba, newPasswd=b'new', style='foo')
        def eb(fail):
            fail.trap(RuntimeError)
            self.assertEqual(fail.getErrorMessage(),
                              "Unknown samba password style 'foo'")
            client.assertNothingSent()
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testPasswordSettingAll_noSamba(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['foo'],
            },
                               complete=1)
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              )
        d.addCallback(cb)
        return d


    def testPasswordSettingAll_hasSamba(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['foo', 'sambaAccount'],
            },
                               complete=1)
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('ntPassword',
                              ['89963F5042E5041A59C249282387A622']),
                delta.Replace('lmPassword',
                              ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d


    def testPasswordSettingAll_hasSambaSam(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['foo', 'sambaSamAccount'],
            },
                               complete=1)
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('sambaNTPassword',
                              ['89963F5042E5041A59C249282387A622']),
                delta.Replace('sambaLMPassword',
                              ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d


    def testPasswordSettingAll_hasSamba_differentCase(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['foo', 'saMBaAccOuNT'],
            },
                               complete=1)
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('ntPassword',
                              ['89963F5042E5041A59C249282387A622']),
                delta.Replace('lmPassword',
                              ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d


    def testPasswordSettingAll_hasSambaSam_differentCase(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['foo', 'sAmbASAmaccoUnt'],
            },
                               complete=1)
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('sambaNTPassword',
                              ['89963F5042E5041A59C249282387A622']),
                delta.Replace('sambaLMPassword',
                              ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP())
        d.addCallback(cb)
        return d


    def testPasswordSettingAll_maybeSamba_WillFind(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [
            pureldap.LDAPSearchResultEntry(objectName='',
                                           attributes=[('objectClass',
                                                        ('foo',
                                                         'sambaAccount',
                                                         'bar'))]),
            pureldap.LDAPSearchResultDone(resultCode=0,
                                          matchedDN='',
                                          errorMessage=''),
            ],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client, dn='cn=foo,dc=example,dc=com')
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(userIdentity='cn=foo,dc=example,dc=com',
                                                   newPasswd=b'new'),
                pureldap.LDAPSearchRequest(baseObject='cn=foo,dc=example,dc=com',
                                           scope=pureldap.LDAP_SCOPE_baseObject,
                                           derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                                           sizeLimit=0,
                                           timeLimit=0,
                                           typesOnly=0,
                                           filter=pureldap.LDAPFilterMatchAll,
                                           attributes=('objectClass',)),
                delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('ntPassword', ['89963F5042E5041A59C249282387A622']),
                delta.Replace('lmPassword', ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX']),
                ]).asLDAP(),
                )
        d.addCallback(cb)
        return d

    def testPasswordSettingAll_maybeSamba_WillNotFind(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [pureldap.LDAPSearchResultEntry(objectName='',
                                            attributes=[('objectClass',
                                                         ('foo',
                                                          'bar'))]),
             pureldap.LDAPSearchResultDone(resultCode=0,
                                          matchedDN='',
                                          errorMessage=''),
            ],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client, dn='cn=foo,dc=example,dc=com')
        d=o.setPassword(newPasswd=b'new')
        def cb(dummy):
            client.assertSent(
                pureldap.LDAPPasswordModifyRequest(userIdentity='cn=foo,dc=example,dc=com',
                                                   newPasswd=b'new'),
                pureldap.LDAPSearchRequest(baseObject='cn=foo,dc=example,dc=com',
                                           scope=pureldap.LDAP_SCOPE_baseObject,
                                           derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                                           sizeLimit=0,
                                           timeLimit=0,
                                           typesOnly=0,
                                           filter=pureldap.LDAPFilterMatchAll,
                                           attributes=('objectClass',)),
                )
        d.addCallback(cb)
        return d

    def testPasswordSettingAll_maybeSamba_WillNotFindAnything(self):
        """LDAPEntry.setPassword(newPasswd=...) changes the password."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=0,
                                           matchedDN='',
                                           errorMessage='')],
            [
             pureldap.LDAPSearchResultDone(resultCode=0,
                                          matchedDN='',
                                          errorMessage=''),
            ],
            [pureldap.LDAPModifyResponse(resultCode=0,
                                         matchedDN='',
                                         errorMessage='')],
            )

        o = ldapsyntax.LDAPEntry(client=client, dn='cn=foo,dc=example,dc=com')
        d = o.setPassword(newPasswd=b'new')

        def checkError(fail):
            fail.trap(ldapsyntax.PasswordSetAggregateError)
            l=fail.value.errors
            assert len(l)==1
            assert len(l[0])==2
            assert l[0][0]=='Samba'
            assert isinstance(l[0][1], failure.Failure)
            l[0][1].trap(ldapsyntax.DNNotPresentError)
            return 'All checks are fine'

        d.addErrback(checkError)

        self.assertEqual('All checks are fine', self.successResultOf(d))
        client.assertSent(
            pureldap.LDAPPasswordModifyRequest(userIdentity='cn=foo,dc=example,dc=com',
                                               newPasswd=b'new'),
            pureldap.LDAPSearchRequest(baseObject='cn=foo,dc=example,dc=com',
                                       scope=pureldap.LDAP_SCOPE_baseObject,
                                       derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                                       sizeLimit=0,
                                       timeLimit=0,
                                       typesOnly=0,
                                       filter=pureldap.LDAPFilterMatchAll,
                                       attributes=('objectClass',)),
            )

    def testPasswordSetting_abortsOnFirstError(self):
        """LDAPEntry.setPassword() aborts on first error (does not parallelize, as it used to)."""
        client = LDAPClientTestDriver(
            [pureldap.LDAPExtendedResponse(resultCode=ldaperrors.LDAPInsufficientAccessRights.resultCode,
                                           matchedDN='',
                                           errorMessage='')],
            )

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['foo', 'sambaAccount'],
            },
                               complete=1)
        d=o.setPassword(newPasswd=b'new')
        def eb(fail):
            fail.trap(ldapsyntax.PasswordSetAggregateError)
            l=fail.value.errors
            assert len(l)==2

            assert len(l[0])==2
            self.assertEqual(l[0][0], 'ExtendedOperation')
            assert isinstance(l[0][1], failure.Failure)
            l[0][1].trap(ldaperrors.LDAPInsufficientAccessRights)

            assert len(l[1])==2
            self.assertEqual(l[1][0], 'Samba')
            assert isinstance(l[1][1], failure.Failure)
            l[1][1].trap(ldapsyntax.PasswordSetAborted)

            client.assertSent(pureldap.LDAPPasswordModifyRequest(
                userIdentity='cn=foo,dc=example,dc=com',
                newPasswd=b'new'),
                              )
        d.addCallbacks(testutil.mustRaise, eb)
        return d


class LDAPSyntaxFetch(unittest.TestCase):
    def testFetch_WithDirtyJournal(self):
        """Trying to fetch attributes with a dirty journal fails."""
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        o['x']=['foo']

        self.failUnlessRaises(
            ldapsyntax.ObjectDirtyError,
            o.fetch)

    def testFetch_Empty(self):
        """Fetching attributes for a newly-created object works."""
        client = LDAPClientTestDriver(
            [   pureldap.LDAPSearchResultEntry(objectName='cn=foo,dc=example,dc=com',
                                               attributes=(
            ('foo', ['a']),
            ('bar', ['b', 'c']),
            )),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),
                ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d=o.fetch()
        def cb(dummy):
            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='cn=foo,dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_baseObject,
                ))

            has=o.keys()
            has.sort()
            want=[b'foo', b'bar']
            want.sort()
            self.assertEqual(has, want)
            self.assertEqual(o['foo'], [b'a'])
            self.assertEqual(o['bar'], [b'b', b'c'])
        d.addCallback(cb)
        return d

    def testFetch_Prefilled(self):
        """Fetching attributes for a (partially) known object overwrites the old attributes."""
        client = LDAPClientTestDriver(
            [   pureldap.LDAPSearchResultEntry(objectName='cn=foo,dc=example,dc=com',
                                               attributes=(
            ('foo', ['a']),
            ('bar', ['b', 'c']),
            )),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),
                ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'foo': ['x'],
            'quux': ['baz', 'xyzzy']
            })
        d=o.fetch()
        def cb(dummy):
            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='cn=foo,dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_baseObject,
                ))

            has=o.keys()
            has.sort()
            want=[b'foo', b'bar']
            want.sort()
            self.assertEqual(has, want)
            self.assertEqual(o['foo'], [b'a'])
            self.assertEqual(o['bar'], [b'b', b'c'])
        d.addCallback(cb)
        return d

    def testFetch_Partial(self):
        """Fetching only some of the attributes does not overwrite existing values of different attribute types."""
        client = LDAPClientTestDriver(
            [   pureldap.LDAPSearchResultEntry(objectName='cn=foo,dc=example,dc=com',
                                               attributes=(
            (b'foo', [b'a']),
            (b'bar', [b'b', b'c']),
            )),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),
                ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            b'foo': [b'x'],
            b'quux': [b'baz', b'xyzzy']
            })
        d=o.fetch(b'foo', b'bar', b'thud')
        def cb(dummy):
            client.assertSent(pureldap.LDAPSearchRequest(
                baseObject='cn=foo,dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_baseObject,
                attributes=(b'foo', b'bar', b'thud'),
                ))

            has=o.keys()
            has.sort()
            want=[b'foo', b'bar', b'quux']
            want.sort()
            self.assertEqual(has, want)
            self.assertEqual(o[b'foo'], [b'a'])
            self.assertEqual(o[b'bar'], [b'b', b'c'])
            self.assertEqual(o[b'quux'], [b'baz', b'xyzzy'])
        d.addCallback(cb)
        return d

    def testCommitAndFetch(self):
        """Fetching after a commit works."""

        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyResponse(resultCode=0,
                                            matchedDN='',
                                            errorMessage='')
                ],
            [   pureldap.LDAPSearchResultEntry('cn=foo,dc=example,dc=com',
                                               [('aValue', ['foo', 'bar'])],
                                               ),
                pureldap.LDAPSearchResultDone(resultCode=0),
                ])
        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            })

        o['aValue']=['foo', 'bar']
        d=o.commit()
        d.addCallback(self.assertIdentical, o)

        d.addCallback(lambda _: o.fetch('aValue'))
        d.addCallback(self.assertIdentical, o)

        def cb(dummy):
            client.assertSent(delta.ModifyOp('cn=foo,dc=example,dc=com', [
                delta.Replace('aValue', ['foo', 'bar']),
                ]).asLDAP(),
                              pureldap.LDAPSearchRequest(
                baseObject='cn=foo,dc=example,dc=com',
                scope=pureldap.LDAP_SCOPE_baseObject,
                attributes=['aValue'],
                ))
        d.addCallback(cb)
        return d

class LDAPSyntaxRDNHandling(unittest.TestCase):
    def testRemovingRDNFails(self):
        """Removing RDN fails with CannotRemoveRDNError."""
        o=ldapsyntax.LDAPEntry(client=None, dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['someObjectClass'],
            'cn': ['foo', 'bar', 'baz'],
            'a': ['aValue'],
            })
        o['cn'].remove('bar')
        del o['a']
        self.assertRaises(ldapsyntax.CannotRemoveRDNError,
                          o['cn'].remove,
                          'foo')
        def f():
            del o['cn']
        self.assertRaises(ldapsyntax.CannotRemoveRDNError,
                          f)
        def f():
            o['cn']=['thud']
        self.assertRaises(ldapsyntax.CannotRemoveRDNError,
                          f)

        # TODO maybe this should be ok, it preserves the RDN.
        # For now, disallow it.
        def f():
            o['cn']=['foo']
        self.assertRaises(ldapsyntax.CannotRemoveRDNError,
                          f)

class LDAPSyntaxMove(unittest.TestCase):
    def test_move(self):
        client = LDAPClientTestDriver(
            [   pureldap.LDAPModifyDNResponse(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com',
                               attributes={
            'objectClass': ['a', 'b'],
            'cn': ['foo'],
            'aValue': ['a'],
            })
        d = o.move('cn=bar,ou=somewhere,dc=example,dc=com')
        def cb(dummy):
            client.assertSent(pureldap.LDAPModifyDNRequest(
                entry='cn=foo,dc=example,dc=com',
                newrdn='cn=bar',
                deleteoldrdn=1,
                newSuperior='ou=somewhere,dc=example,dc=com',
                ))

            self.assertEqual(o.dn, u'cn=bar,ou=somewhere,dc=example,dc=com')
        d.addCallback(cb)
        return d

class Bind(unittest.TestCase):
    def test_ok(self):
        client = LDAPClientTestDriver(
            [   pureldap.LDAPBindResponse(resultCode=0,
                                          matchedDN=''),
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d = defer.maybeDeferred(o.bind, 's3krit')
        d.addCallback(self.assertIdentical, o)
        def cb(dummy):
            client.assertSent(pureldap.LDAPBindRequest(
                dn='cn=foo,dc=example,dc=com',
                auth='s3krit'))
        d.addCallback(cb)
        return d

    def test_fail(self):
        client = LDAPClientTestDriver(
            [   pureldap.LDAPBindResponse(
            resultCode=ldaperrors.LDAPInvalidCredentials.resultCode,
            matchedDN=''),
                ])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d = defer.maybeDeferred(o.bind, 's3krit')
        def eb(fail):
            fail.trap(ldaperrors.LDAPInvalidCredentials)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_err(self):
        client = LDAPClientTestDriver([
                failure.Failure(error.ConnectionLost())])

        o=ldapsyntax.LDAPEntry(client=client,
                               dn='cn=foo,dc=example,dc=com')
        d = defer.maybeDeferred(o.bind, 'whatever')
        def eb(fail):
            fail.trap(error.ConnectionLost)
        d.addCallbacks(testutil.mustRaise, eb)
        return d



"""
Test cases for ldaptor.entry
"""

from twisted.trial import unittest

from ldaptor import delta, entry
from ldaptor.protocols.ldap.ldaperrors import LDAPInvalidCredentials



class TestBaseLDAPEntry(unittest.TestCase):
    """
    Tests for ldaptor.entry.BaseLDAPEntry.
    """

    def testEqualitySameType(self):
        """
        It is equal if it has the same DN (case insensitive)
        and same attributes and values.
        """
        a = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(
            dn='Dc=Foo',
            attributes={
                'foo': ['bar'],
            })

        self.assertEqual(a, b)

    def testEqualityDifferentType(self):
        """
        It is not equal with objects of different types.
        """
        a = entry.BaseLDAPEntry(dn='dc=foo', attributes={})
        self.assertFalse(a == object())

    def testInequalityDifferentDN(self):
        """
        Entries are not equal if their DNs are not equal
        """
        a = entry.BaseLDAPEntry(dn='dn=foo', attributes={'foo': ['bar']})
        b = entry.BaseLDAPEntry(dn='dn=bar', attributes={'foo': ['bar']})
        self.assertNotEqual(a, b)

    def testBindPlainText(self):
        """
        It will bind when the password for the entry is stored in plain text,
        and will return a deferred which has itself as callback.
        """
        sut = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'userPassword': [b'some-plain-text'],
            })

        deferred = sut.bind(b'some-plain-text')
        result = self.successResultOf(deferred)

        self.assertIs(sut, result)


    def testBindSeededSHA(self):
        """
        It can bind with password stored in seeded SHA.
        """
        sut = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'userPassword': [b'{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8='],
            })

        deferred = sut.bind(b'secret')
        result = self.successResultOf(deferred)

        self.assertIs(sut, result)


    def testBindPlainTextError(self):
        """
        Return a LDAPInvalidCredentials failure when password don't match.
        """
        sut = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'userPassword': [b'some-plain-text'],
            })

        deferred = sut.bind(b'other-password')
        failure = self.failureResultOf(deferred)

        self.assertTrue(failure.check(LDAPInvalidCredentials))



    def testBindSHAError(self):
        """
        Return a LDAPInvalidCredentials failure when encoded password don't
        match.
        """
        sut = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'userPassword': [b'{SSHA}anythinghere'],
            })

        deferred = sut.bind(b'other-password')
        failure = self.failureResultOf(deferred)

        self.assertTrue(failure.check(LDAPInvalidCredentials))

    def testGetLDIF(self):
        """
        Getting human readable representation of an entry
        """
        sut = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'foo': ['bar'],
                'bar': ['foo'],
            }
        )
        self.assertEqual(sut.getLDIF(), u'dn: dc=foo\nbar: foo\nfoo: bar\n\n')

    def testNonzero(self):
        """Entry is always non-zero"""
        sut = entry.BaseLDAPEntry(dn='')
        self.assertTrue(bool(sut))

    def testRepr(self):
        """
        Getting string representation of an entry
        """
        sut = entry.BaseLDAPEntry(
            dn='dc=foo',
            attributes={
                'foo': ['bar'],
                'bar': ['foo'],
            }
        )
        self.assertEqual(repr(sut), "BaseLDAPEntry('dc=foo', {'bar': ['foo'], 'foo': ['bar']})")



class TestDiffEntry(unittest.TestCase):
    """
    Tests for ldaptor.entry.BaseLDAPEntry.diff()
    """
    def testEqual(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        result = a.diff(b)
        self.assertEqual(result, None)

    def testAdd_New_OneType_OneValue(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['quux'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Add('baz', ['quux']),
            ]))

    def testAdd_New_OneType_ManyValues(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['quux', 'thud', 'foo'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Add('baz', ['quux', 'thud', 'foo']),
            ]))

    def testAdd_New_ManyTypes(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['quux'],
            'bang': ['thud'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Add('bang', ['thud']),
            delta.Add('baz', ['quux']),
            ]))

    def testAdd_Existing_OneType_OneValue(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar', 'quux'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Add('foo', ['quux']),
            ]))

    def testAdd_Existing_OneType_ManyValues(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar', 'quux', 'thud', 'foo'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Add('foo', ['quux', 'thud', 'foo']),
            ]))

    def testAdd_NewAndExisting_ManyTypes(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['quux'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar', 'thud', 'bang'],
            'baz': ['quux', 'bar', 'stump'],
            'bang': ['thud', 'barble'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Add('bang', ['thud', 'barble']),
            delta.Add('baz', ['bar', 'stump']),
            delta.Add('foo', ['thud', 'bang']),
            ]))

    def testDelete_All_OneType(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['quux', 'thud'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Delete('baz', ['quux', 'thud']),
            ]))

    def testDelete_Some_OneType(self):
        a = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['quux', 'thud'],
            })
        b = entry.BaseLDAPEntry(dn='dc=foo',
                                attributes={
            'foo': ['bar'],
            'baz': ['thud'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('dc=foo',
                                         [
            delta.Delete('baz', ['quux']),
            ]))

    def testComplex(self):
        a = entry.BaseLDAPEntry(dn='cn=Paula Jensen,ou=Product Development,dc=airius,dc=com',
                                attributes={
            'description': ['Something'],
            'telephonenumber': ['+123 456'],
            'facsimiletelephonenumber': ['+1 408 555 9876'],
            })
        b = entry.BaseLDAPEntry(dn='cn=Paula Jensen,ou=Product Development,dc=airius,dc=com',
                                attributes={
            'postalAddress': ['123 Anystreet $ Sunnyvale, CA $ 94086'],
            'telephonenumber': ['+1 408 555 1234', '+1 408 555 5678'],
            })
        result = a.diff(b)
        self.assertEqual(result,
                          delta.ModifyOp('cn=Paula Jensen,ou=Product Development,dc=airius,dc=com',
                                         [
            delta.Add('postalAddress', ['123 Anystreet $ Sunnyvale, CA $ 94086']),
            delta.Delete('description', ['Something']),
            delta.Delete('facsimiletelephonenumber', ['+1 408 555 9876']),
            delta.Add('telephonenumber', ['+1 408 555 1234', '+1 408 555 5678']),
            delta.Delete('telephonenumber', ['+123 456']),
            ]))



from ldaptor.protocols.ldap import ldaperrors, ldapsyntax
from ldaptor.protocols import pureldap
from ldaptor import schema
from ldaptor._encoder import to_bytes


def _fetchCb(subschemaSubentry, client):
    o = ldapsyntax.LDAPEntry(client=client,
                             dn=subschemaSubentry)
    d = o.search(scope=pureldap.LDAP_SCOPE_baseObject,
                 sizeLimit=1,
                 attributes=["attributeTypes", "objectClasses"])

    def handleSearchResults(l):
        if len(l) == 0:
            raise ldaperrors.LDAPOther("No such DN")
        elif len(l) == 1:
            o = l[0]

            attributeTypes = []
            objectClasses = []
            for text in o.get("attributeTypes", []):
                attributeTypes.append(schema.AttributeTypeDescription(to_bytes(text)))
            for text in o.get("objectClasses", []):
                objectClasses.append(schema.ObjectClassDescription(to_bytes(text)))
            assert attributeTypes, "LDAP server doesn't give attributeTypes for subschemaSubentry dn=%s" % o.dn
            return (attributeTypes, objectClasses)
        else:
            raise ldaperrors.LDAPOther("DN matched multiple entries")

    d.addCallback(handleSearchResults)
    return d


def fetch(client, baseObject):
    o = ldapsyntax.LDAPEntry(client=client,
                             dn=baseObject)
    d = o.search(scope=pureldap.LDAP_SCOPE_baseObject,
                 sizeLimit=1,
                 attributes=["subschemaSubentry"])

    def handleSearchResults(l):
        if len(l) == 0:
            raise ldaperrors.LDAPOther("No such DN")
        elif len(l) == 1:
            o = l[0]
            assert "subschemaSubentry" in o, "No subschemaSubentry. TODO"
            subSchemas = o["subschemaSubentry"]
            assert len(subSchemas) == 1, "More than one subschemaSubentry is not support yet. TODO"
            for s in subSchemas:
                return s
        else:
            raise ldaperrors.LDAPOther("DN matched multiple entries")

    d.addCallback(handleSearchResults)
    d.addCallback(_fetchCb, client)
    return d



from passlib.hash import (nthash as passlib_nthash, lmhash as passlib_lmhash)

from ldaptor import config

def nthash(password=b''):
    """Generates nt md4 password hash for a given password."""
    return passlib_nthash.hash(password[:128]).encode('ascii').upper()


def lmhash_locked(password=b''):
    """
    Generates a lanman password hash that matches no password.

    Note that the author thinks LanMan hashes should be banished from
    the face of the earth.
    """
    return 32 * b'X'


def lmhash(password=b''):
    """
    Generates lanman password hash for a given password.

    Note that the author thinks LanMan hashes should be banished from
    the face of the earth.
    """

    if not config.useLMhash():
        return lmhash_locked()

    return passlib_lmhash.hash(password).encode('ascii').upper()



"""
Test cases for ldaptor.protocols.ldap.delta
"""

from twisted.trial import unittest
from ldaptor import delta, entry, attributeset, inmemory
from ldaptor.protocols.ldap import ldapsyntax, distinguishedname, ldaperrors

class TestModifications(unittest.TestCase):
    def setUp(self):
        self.foo = ldapsyntax.LDAPEntry(
            None,
            dn='cn=foo,dc=example,dc=com',
            attributes={
            'objectClass': ['person'],
            'cn': ['foo', 'thud'],
            'sn': ['bar'],
            'more': ['junk'],
            })

    def testAddOld(self):
        mod = delta.Add('cn', ['quux'])
        mod.patch(self.foo)

        self.failIf('stuff' in self.foo)
        self.failUnlessEqual(self.foo['cn'], ['foo', 'thud', 'quux'])

    def testAddNew(self):
        mod = delta.Add('stuff', ['val1', 'val2'])
        mod.patch(self.foo)

        self.failUnlessEqual(self.foo['stuff'], ['val1', 'val2'])
        self.failUnlessEqual(self.foo['cn'], ['foo', 'thud'])

    def testDelete(self):
        mod = delta.Delete('cn', ['thud'])
        mod.patch(self.foo)

        self.failIf('stuff' in self.foo)
        self.failUnlessEqual(self.foo['cn'], ['foo'])

    def testDeleteAll(self):
        mod = delta.Delete('more')
        mod.patch(self.foo)

        self.failIf('stuff' in self.foo)
        self.failUnlessEqual(self.foo['cn'], ['foo', 'thud'])

    def testDelete_FailOnNonExistingAttributeType_All(self):
        mod = delta.Delete('notexist', [])
        self.assertRaises(KeyError,
                          mod.patch,
                          self.foo)

    def testDelete_FailOnNonExistingAttributeType_OneValue(self):
        mod = delta.Delete('notexist', ['a'])
        self.assertRaises(KeyError,
                          mod.patch,
                          self.foo)

    def testDelete_FailOnNonExistingAttributeValue(self):
        mod = delta.Delete('cn', ['notexist'])
        self.assertRaises(LookupError,
                          mod.patch,
                          self.foo)


    def testReplace_Add(self):
        mod = delta.Replace('stuff', ['val1', 'val2'])
        mod.patch(self.foo)

        self.failUnlessEqual(self.foo['stuff'], ['val1', 'val2'])
        self.failUnlessEqual(self.foo['sn'], ['bar'])
        self.failUnlessEqual(self.foo['more'], ['junk'])

    def testReplace_Modify(self):
        mod = delta.Replace('sn', ['baz'])
        mod.patch(self.foo)

        self.failIf('stuff' in self.foo)
        self.failUnlessEqual(self.foo['sn'], ['baz'])
        self.failUnlessEqual(self.foo['more'], ['junk'])

    def testReplace_Delete_Existing(self):
        mod = delta.Replace('more', [])
        mod.patch(self.foo)

        self.failIf('stuff' in self.foo)
        self.failUnlessEqual(self.foo['sn'], ['bar'])
        self.failIf('more' in self.foo)

    def testReplace_Delete_NonExisting(self):
        mod = delta.Replace('nonExisting', [])
        mod.patch(self.foo)

        self.failIf('stuff' in self.foo)
        self.failUnlessEqual(self.foo['sn'], ['bar'])
        self.failUnlessEqual(self.foo['more'], ['junk'])



class TestModificationOpLDIF(unittest.TestCase):
    def testAdd(self):
        m=delta.Add('foo', ['bar', 'baz'])
        self.assertEqual(m.asLDIF(),
                         b"""\
add: foo
foo: bar
foo: baz
-
""")

    def testDelete(self):
        m=delta.Delete('foo', ['bar', 'baz'])
        self.assertEqual(m.asLDIF(),
                         b"""\
delete: foo
foo: bar
foo: baz
-
""")

    def testDeleteAll(self):
        m=delta.Delete('foo')
        self.assertEqual(m.asLDIF(),
                         b"""\
delete: foo
-
""")

    def testReplace(self):
        m=delta.Replace('foo', ['bar', 'baz'])
        self.assertEqual(m.asLDIF(),
                         b"""\
replace: foo
foo: bar
foo: baz
-
""")

    def testReplaceAll(self):
        m=delta.Replace('thud')
        self.assertEqual(m.asLDIF(),
                         b"""\
replace: thud
-
""")

    def testAddBase64(self):
        """
        LDIF attribute representation is base64 encoded
        if attribute value contains nonprintable characters
        or starts with reserved characters
        """
        m = delta.Add('attr', [':value1', 'value\n\r2'])
        self.assertEqual(m.asLDIF(),
                         b"""\
add: attr
attr:: OnZhbHVlMQ==
attr:: dmFsdWUKDTI=
-
""")


class OperationTestCase(unittest.TestCase):
    """
    Test case for operations on a LDAP tree.
    """
    def getRoot(self):
        """
        Returns a new LDAP root for dc=example,dc=com.
        """
        return inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))


class TestAddOpLDIF(OperationTestCase):
    """
    Unit tests for `AddOp`.
    """
    def testAsLDIF(self):
        """
        It will return the LDIF representation of the operation.
        """
        sut =delta.AddOp(entry.BaseLDAPEntry(
            dn='dc=example,dc=com',
            attributes={
                'foo': ['bar', 'baz'],
                'quux': ['thud'],
                },
            ))

        result = sut.asLDIF()

        self.assertEqual(b"""dn: dc=example,dc=com
changetype: add
foo: bar
foo: baz
quux: thud

""",
        result)

    def testAddOpEqualitySameEntry(self):
        """
        Objects are equal when the have the same LDAP entry.
        """
        first_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['same', 'attributes']})
        second_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['same', 'attributes']})

        first = delta.AddOp(first_entry)
        second = delta.AddOp(second_entry)

        self.assertEqual(first, second)

    def testAddOpInequalityDifferentEntry(self):
        """
        Objects are not equal when the have different LDAP entries.
        """
        first_entry = entry.BaseLDAPEntry(
            dn='ou=First Team, dc=example,dc=com',
            attributes={'foo': ['same', 'attributes']})
        second_entry = entry.BaseLDAPEntry(
            dn='ou=First Team, dc=example,dc=com',
            attributes={'foo': ['other', 'attributes']})

        first = delta.AddOp(first_entry)
        second = delta.AddOp(second_entry)

        self.assertNotEqual(first, second)

    def testAddOpInequalityNoEntryObject(self):
        """
        Objects is not equal with random objects.
        """
        team_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['same', 'attributes']})
        sut = delta.AddOp(team_entry)

        self.assertNotEqual(sut, {'foo': ['same', 'attributes']})

    def testAddOpHashSimilar(self):
        """
        Objects which are equal have the same hash.
        """
        first_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['same', 'attributes']})
        second_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['same', 'attributes']})

        first = delta.AddOp(first_entry)
        second = delta.AddOp(second_entry)

        self.assertEqual(hash(first), hash(second))

    def testAddOpHashDifferent(self):
        """
        Objects which are not equal have different hash.
        """
        first_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['one', 'attributes']})
        second_entry = entry.BaseLDAPEntry(
            dn='ou=Duplicate Team, dc=example,dc=com',
            attributes={'foo': ['other', 'attributes']})

        first = delta.AddOp(first_entry)
        second = delta.AddOp(second_entry)

        self.assertNotEqual(hash(first), hash(second))

    def testAddOp_DNExists(self):
        """
        It fails to perform the `add` operation for an existing entry.
        """
        root = self.getRoot()
        root.addChild(
            rdn='ou=Existing Team',
            attributes={
            'objectClass': ['a', 'b'],
            'ou': ['HR'],
            })

        hr_entry = entry.BaseLDAPEntry(
            dn='ou=Existing Team, dc=example,dc=com',
            attributes={'foo': ['dont', 'care']})
        sut = delta.AddOp(hr_entry)

        deferred = sut.patch(root)

        failure = self.failureResultOf(deferred)
        self.assertIsInstance(failure.value, ldaperrors.LDAPEntryAlreadyExists)

    def testRepr(self):
        """
        Getting string representation
        """
        sut = delta.AddOp(entry.BaseLDAPEntry(
            dn='dc=example,dc=com',
            attributes={
                'bar': ['foo'],
                'foo': ['bar'],
            },
        ))

        self.assertEqual(repr(sut), "AddOp(BaseLDAPEntry('dc=example,dc=com', "
                                    "{'bar': ['foo'], 'foo': ['bar']}))")


class TestDeleteOpLDIF(OperationTestCase):
    """
    Unit tests for DeleteOp.
    """
    def testAsLDIF(self):
        """
        It return the LDIF representation of the delete operation.
        """
        sut = delta.DeleteOp('dc=example,dc=com')

        result = sut.asLDIF()
        self.assertEqual(b"""dn: dc=example,dc=com
changetype: delete

""",
        result)

    def testDeleteOpEqualitySameDN(self):
        """
        Objects are equal when the have the same DN.
        """
        first_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')
        second_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')

        first = delta.DeleteOp(first_entry)
        second = delta.DeleteOp(second_entry)

        self.assertEqual(first, second)

    def testDeleteOpEqualityEqualDN(self):
        """
        DeleteOp objects are equal if their DNs are equal.
        """
        first_dn = distinguishedname.DistinguishedName(stringValue='ou=Team,dc=example,dc=com')
        first = delta.DeleteOp(first_dn)

        second_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example, dc=com')
        second = delta.DeleteOp(second_entry)

        third = delta.DeleteOp('ou=Team, dc=example,dc=com')

        self.assertEqual(first, second)
        self.assertEqual(first, third)

    def testDeleteOpInequalityDifferentEntry(self):
        """
        DeleteOp objects are not equal when the have different LDAP entries.
        """
        first_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')
        second_entry = entry.BaseLDAPEntry(dn='ou=Cowboys, dc=example,dc=com')

        first = delta.DeleteOp(first_entry)
        second = delta.DeleteOp(second_entry)

        self.assertNotEqual(first, second)

    def testDeleteOpInequalityNoEntryObject(self):
        """
        DeleteOp objects is not equal with random objects.
        """
        team_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')

        sut = delta.DeleteOp(team_entry)

        self.assertNotEqual(sut, 'ou=Team, dc=example,dc=com')

    def testDeleteOpHashSimilar(self):
        """
        Objects which are equal have the same hash.
        """
        first_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')
        second_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')

        first = delta.DeleteOp(first_entry)
        second = delta.DeleteOp(second_entry)

        self.assertEqual(hash(first), hash(second))

    def testDeleteOpHashDifferent(self):
        """
        Objects which are not equal have different hash.
        """
        first_entry = entry.BaseLDAPEntry(dn='ou=Team, dc=example,dc=com')
        second_entry = entry.BaseLDAPEntry(dn='ou=Cowboys, dc=example,dc=com')

        first = delta.DeleteOp(first_entry)
        second = delta.DeleteOp(second_entry)

        self.assertNotEqual(hash(first), hash(second))

    def testDeleteOp_DNNotFound(self):
        """
        If fail to delete when the RDN does not exists.
        """
        root = self.getRoot()
        sut = delta.DeleteOp('cn=nope,dc=example,dc=com')

        deferred = sut.patch(root)

        failure = self.failureResultOf(deferred)
        self.assertIsInstance(failure.value, ldaperrors.LDAPNoSuchObject)

    def testDeleteOpInvalidDN(self):
        """
        Invalid type of DN raises AssertionError
        """
        self.assertRaises(AssertionError, delta.DeleteOp, 0)

    def testRepr(self):
        """
        Getting string representation
        """
        sut = delta.DeleteOp('dc=example,dc=com')

        self.assertEqual(repr(sut), "DeleteOp('dc=example,dc=com')")


class TestModifyOp(OperationTestCase):
    """
    Unit tests for ModifyOp.
    """

    def testAsLDIF(self):
        """
        It will return a LDIF representation of the contained operations.
        """
        sut = delta.ModifyOp(
            'cn=Paula Jensen, ou=Dev Ops, dc=airius, dc=com',
            [
                delta.Add(
                    'postaladdress',
                    ['123 Anystreet $ Sunnyvale, CA $ 94086'],
                    ),
                delta.Delete('description'),
                delta.Replace(
                    'telephonenumber',
                    ['+1 408 555 1234', '+1 408 555 5678'],
                    ),
                delta.Delete(
                    'facsimiletelephonenumber', ['+1 408 555 9876']),
                ]
            )

        result = sut.asLDIF()

        self.assertEqual(b"""dn: cn=Paula Jensen,ou=Dev Ops,dc=airius,dc=com
changetype: modify
add: postaladdress
postaladdress: 123 Anystreet $ Sunnyvale, CA $ 94086
-
delete: description
-
replace: telephonenumber
telephonenumber: +1 408 555 1234
telephonenumber: +1 408 555 5678
-
delete: facsimiletelephonenumber
facsimiletelephonenumber: +1 408 555 9876
-

""",
        result,
        )

    def testInequalityDiffertnDN(self):
        """
        Modify operations for different DN are not equal.
        """
        first = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
            )

        second = delta.ModifyOp(
            'cn=doe,dc=example,dc=com',
            [delta.Delete('description')]
            )

        self.assertNotEqual(first, second)

    def testInequalityDifferentModifications(self):
        """
        Modify operations with different modifications are not equal
        """
        first = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Add('description')]
        )

        second = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
        )

        self.assertNotEqual(first, second)

    def testInequalityNotModifyOP(self):
        """
        Modify operations are not equal with other object types.
        """
        sut = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
            )

        self.assertNotEqual('cn=john,dc=example,dc=com', sut)

    def testInequalityDiffertnOperations(self):
        """
        Modify operations for same DN but different operations are not equal.
        """
        first = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
            )
        second = delta.ModifyOp(
            'cn=doe,dc=example,dc=com',
            [delta.Delete('homeDirectory')]
            )

        self.assertNotEqual(first, second)

    def testHashEquality(self):
        """
        Modify operations can be hashed and equal objects have the same
        hash.
        """
        first = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
            )

        second = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
            )

        self.assertEqual(first, second)
        self.assertEqual(
            first.asLDIF(), second.asLDIF(),
            'LDIF equality is a precondition for valid hash values',
            )
        self.assertEqual(hash(first), hash(second))

    def testHashInequality(self):
        """
        Different modify operations have different hash values.
        """
        first = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
            )

        second = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('homeDirectory')]
            )

        self.assertNotEqual(first.asLDIF(), second.asLDIF())
        self.assertNotEqual(hash(first), hash(second))

    def testModifyOp_DNNotFound(self):
        """
        If fail to modify when the RDN does not exists.
        """
        root = self.getRoot()
        sut = delta.ModifyOp(
            'cn=nope,dc=example,dc=com',
            [delta.Add('foo', ['bar'])],
            )

        deferred = sut.patch(root)

        failure = self.failureResultOf(deferred)
        self.assertIsInstance(failure.value, ldaperrors.LDAPNoSuchObject)

    def testRepr(self):
        """
        Getting string representation
        """
        sut = delta.ModifyOp(
            'cn=john,dc=example,dc=com',
            [delta.Delete('description')]
        )

        self.assertEqual(repr(sut), "ModifyOp(dn='cn=john,dc=example,dc=com', "
                                    "modifications=[Delete('description', [])])")


class TestModificationComparison(unittest.TestCase):
    def testEquality_Add_True(self):
        a = delta.Add('k', ['b', 'c', 'd'])
        b = delta.Add('k', ['b', 'c', 'd'])
        self.assertEqual(a, b)

    def testEquality_AddVsDelete_False(self):
        a = delta.Add('k', ['b', 'c', 'd'])
        b = delta.Delete('k', ['b', 'c', 'd'])
        self.assertNotEquals(a, b)

    def testEquality_AttributeSet_False(self):
        a = delta.Add('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        self.assertNotEquals(a, b)

    def testEquality_List_False(self):
        a = delta.Add('k', ['b', 'c', 'd'])
        b = ['b', 'c', 'd']
        self.assertNotEquals(a, b)

